from typing import Any

from . import boolean


def flag_string(value: str) -> Any:
    """
    any_string => skip
    !type_name:real_value => type_cls(value)
    :param value:
    :return:
    """
    if value.startswith("!"):
        type_name, real_val = value[1:].split(":", maxsplit=1)
        cls = eval(type_name)
        if type(cls) is not type:
            raise ValueError(f"{type_name} is not a type")
        if cls is bool:
            return boolean(real_val)
        return cls(real_val)
    return value



from argparse import Namespace
from pathlib import Path
from typing import Dict, TYPE_CHECKING

import yaml
from abc import ABC, abstractmethod

if TYPE_CHECKING:
    from ..config import BaseConfig


class BaseConfigLoader(ABC):
    @abstractmethod
    def load(self, config: 'BaseConfig') -> 'BaseConfig':
        pass


class DictBasedConfigLoader(BaseConfigLoader):

    @abstractmethod
    def get_dict(self) -> Dict:
        pass

    def load(self, config: 'BaseConfig'):
        for key, value in self.get_dict().items():
            config[key] = value


class PathBasedConfigLoader(BaseConfigLoader, ABC):
    _path: Path

    @property
    def path(self) -> Path:
        return self._path

    @path.setter
    def path(self, path: Path) -> None:
        self._path = path


class YamlConfigLoader(DictBasedConfigLoader, PathBasedConfigLoader):

    def get_dict(self) -> Dict:
        if not self.path.is_file():
            raise FileNotFoundError(str(self.path))
        stream = self.path.open()
        data = yaml.safe_load(stream)

        if data is None:
            data = {}

        stream.close()
        return data


class DictConfigLoader(DictBasedConfigLoader):
    _dict: Dict

    def __init__(self, _dict: Dict):
        self._dict = _dict

    def get_dict(self) -> Dict:
        return self._dict


class NamespaceConfigLoader(DictBasedConfigLoader):
    _args: Namespace

    def __init__(self, args: Namespace):
        self._args = args

    def get_dict(self) -> Dict:
        return vars(self._args)



from typing import Dict, Type

from .. import config as cfg
from fancy.config import exceptions, utils


class ConfigLoaderFactory:
    suffix: str = "ConfigLoader"
    loaders = None

    @classmethod
    def create_loader(cls, method: str = 'yaml') -> cfg.BaseConfigLoader:
        loaders = cls.get_all_loaders()
        if method + cls.suffix in loaders:
            method += cls.suffix
        elif method.capitalize() + cls.suffix in loaders:
            method = method.capitalize() + cls.suffix
        elif method not in loaders:
            raise exceptions.ClassNotFoundException(method)
        return loaders[method]()

    @classmethod
    def get_all_loaders(cls) -> Dict[str, Type[cfg.BaseConfigLoader]]:
        if cls.loaders is None:
            cls.loaders = {
                loader.__name__: loader for loader in utils.reflections.find_subclasses(cfg.BaseConfigLoader)
            }
        return cls.loaders

    @classmethod
    def get_suffix(cls) -> str:
        return cls.suffix

    @classmethod
    def set_suffix(cls, suffix: str) -> None:
        cls.suffix = suffix



__all__ = [
    "boolean", "config", "flag_string", "flag_container"
]

from .boolean import boolean
from .config import config
from .flag_string import flag_string
from .flag_container import flag_container



from typing import Any

TRUE_CHOICES = {"on", "1", "true", "yes", "==TRUE=="}
FALSE_CHOICES = {"off", "0", "false", "no", "==FALSE=="}


def boolean(value: Any) -> bool:
    if isinstance(value, str):
        _value = value.lower().strip()
        if _value in TRUE_CHOICES:
            return True
        if _value in FALSE_CHOICES:
            return False
        raise ValueError(
            f"unexpected value: {value}, expected choices:"
            f" {str(TRUE_CHOICES.union(FALSE_CHOICES))}"
        )
    return bool(value)



import warnings
from pathlib import Path
from typing import Any, List


def identical(val: Any) -> Any:
    return val


def to_string(val: Any) -> str:
    warnings.warn("to_string has deprecated.use str to instead.", DeprecationWarning)
    return str(val)


def to_integer(val: Any) -> int:
    warnings.warn("to_integer has deprecated.use int to instead.", DeprecationWarning)
    return int(val)


def to_float(val: Any) -> float:
    warnings.warn("to_float has deprecated.use float to instead.", DeprecationWarning)
    return float(val)


def to_bool(val: Any) -> bool:
    warnings.warn("to_bool has deprecated.use bool to instead.", DeprecationWarning)
    return bool(val)


def to_path(val: Any) -> Path:
    warnings.warn("to_path has deprecated.use pathlib.Path to instead.", DeprecationWarning)
    return Path(val)


def to_list(val: Any) -> List:
    warnings.warn("to_list has deprecated.use list to instead.", DeprecationWarning)
    return list(val)



from typing import Type, Callable, TYPE_CHECKING


if TYPE_CHECKING:
    from .. import BaseConfig


def config(config_type: Type["BaseConfig"]) -> Callable:
    from .. import DictConfigLoader

    def _inner(config_dict: dict) -> "BaseConfig":
        return config_type(DictConfigLoader(config_dict))  # lazy import to avoid circular import
    return _inner



__all__ = [
    'BaseConfig', 'Option',
    'identical', 'to_string', 'to_integer', 'to_float', 'to_bool', 'to_path', 'to_list',
    'BaseConfigLoader', 'DictBasedConfigLoader', 'YamlConfigLoader', 'DictConfigLoader', 'PathBasedConfigLoader',
    'ConfigLoaderFactory', 'ConfigFactory',
    'process'
]
from .option_preprocessing import (
    identical,
    to_float,
    to_integer,
    to_string,
    to_bool,
    to_path,
    to_list
)
from . import process
from .option import Option
from .config import BaseConfig
from .config_loaders import (
    BaseConfigLoader,
    DictBasedConfigLoader,
    PathBasedConfigLoader,
    YamlConfigLoader,
    DictConfigLoader
)


from .config_loader_factory import ConfigLoaderFactory
from .config_factory import ConfigFactory



from abc import ABC
from typing import TYPE_CHECKING, Dict, List

from ..config import Option

if TYPE_CHECKING:
    from ..config import BaseConfigLoader


class BaseConfig(ABC):
    _name_mapping: Dict[str, str]
    _all_options: Dict[str, Option]
    _all_required_options: List[Option]

    def __init__(self, loader: 'BaseConfigLoader'):
        loader.load(self)
        for option in self.get_all_required_options():
            if not hasattr(self, option.name):
                raise ValueError(f'the missing option {option.name} is required.')
        self.post_load()

    def __getitem__(self, item):
        if isinstance(item, str):
            raise TypeError(f'index must be str, not {type(item)}')
        try:
            return self.__getattribute__(self.get_name_mapping()[item])
        except AttributeError:
            raise IndexError(f'not contains the config named {item}')

    def __setitem__(self, key, value):
        if not isinstance(key, str):
            raise TypeError(f'index must be str, not {type(key)}')
        key = self.get_name_mapping()[key]
        if key not in type(self).__dict__:
            raise IndexError(f'not contains the config named {key}')
        self.__setattr__(key, value)

    def post_load(self):
        pass

    @classmethod
    def get_all_options(cls) -> Dict[str, Option]:
        if not hasattr(cls, "_all_options"):
            cls._all_options = {name: option for name, option in vars(cls).items() if isinstance(option, Option)}
        return cls._all_options

    @classmethod
    def get_all_required_options(cls) -> List[Option]:
        if not hasattr(cls, "_all_required_options"):
            cls._all_required_options = [option for option in cls.get_all_options().values() if option.required]
        return cls._all_required_options

    @classmethod
    def get_name_mapping(cls) -> Dict[str, str]:
        if not hasattr(cls, "_name_mapping"):
            cls._name_mapping = {option.name: attr_name for attr_name, option in cls.get_all_options().items()}
        return cls._name_mapping



from typing import List


def find_subclasses(cls: type) -> List[type]:
    cls_list = []
    search = [cls]
    while len(search) > 0:
        sub_cls = search.pop()
        cls_list.append(sub_cls)
        search += sub_cls.__subclasses__()
    return cls_list



from typing import Dict, Type, TYPE_CHECKING

from . import BaseConfig, utils
from .exceptions import ClassNotFoundException


if TYPE_CHECKING:
    from . import BaseConfigLoader


class ConfigFactory:
    configs = None
    suffix: str = "Config"

    @classmethod
    def create_config(cls, config: str, loader: 'BaseConfigLoader') -> 'BaseConfig':
        configs = cls._get_all_configs()
        if config + cls.suffix in configs:
            config += cls.suffix
        elif config[0].capitalize() + config[1:] + cls.suffix in configs:
            config = config[0].capitalize() + config[1:] + cls.suffix
        elif config not in configs:
            raise ClassNotFoundException(config)
        return configs[config](loader)

    @classmethod
    def _get_all_configs(cls) -> Dict[str, Type['BaseConfig']]:
        if cls.configs is None:
            cls.configs = {config.__name__: config for config in utils.reflections.find_subclasses(BaseConfig)}
        return cls.configs

    @classmethod
    def get_suffix(cls) -> str:
        return cls.suffix

    @classmethod
    def set_suffix(cls, suffix: str) -> None:
        cls.suffix = suffix



from pathlib import Path

from setuptools import setup, find_namespace_packages

package_dir = Path("./fancy/config")
readme_file = package_dir / "README.md"
requirements_file = package_dir.parent.parent / "requirements.txt"

with readme_file.open() as fp:
    long_description = fp.read()

with requirements_file.open() as fp:
    requirements = [r.strip() for r in fp.readlines()]


setup(
    name="fancy-config",
    version="0.4.0.a0",
    packages=find_namespace_packages(),
    package_data={
        # If any package contains *.txt or *.rst files, include them:
        "": ["*.md", "*.txt"],
    },

    # metadata to display on PyPI
    author="su-rin",
    author_email="susautw@gmail.com",
    description="config",
    license="MIT",
    keywords="config",
    project_urls={
        "Source Code": "https://github.com/susautw/fancy-config",
    },
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=requirements
)



from abc import ABC, abstractmethod
from typing import Any, Callable


class BaseOptionConverter(ABC):
    @abstractmethod
    def set_name(self, name: str):
        pass

    @abstractmethod
    def set_required(self, required: bool):
        pass

    @abstractmethod
    def set_nullable(self, nullable: bool):
        pass

    @abstractmethod
    def set_default(self, default: Any):
        pass

    @abstractmethod
    def set_type(self, type: Callable[[Any], Any]):
        pass

    @abstractmethod
    def set_description(self, description: str):
        pass



import warnings
from typing import Any, Callable, Type, TYPE_CHECKING

from .process import boolean, config
from ..config import identical

if TYPE_CHECKING:
    from ..config import BaseConfig


class Option:
    _type: Callable[[Any], Any]
    _required: bool
    _nullable: bool
    _default: Any

    _description: str
    _deleted_suffix: str = "__deleted"

    auto_boolean_process: bool
    auto_config_process: bool

    def __init__(
            self,
            required=False,
            nullable=False, default=None,
            option_type=object,
            type=identical,
            preprocess=identical,
            description="",
            name: str = None,
            auto_boolean_process: bool = True,
            auto_config_process: bool = True
    ):
        self.__name__ = name
        self._required = required
        self._nullable = nullable
        self._type = option_type
        self._default = default
        self._description = description
        self.auto_boolean_process = auto_boolean_process
        self.auto_config_process = auto_config_process

        if preprocess is not identical:
            warnings.warn("preprocess has deprecated. use type to instead.", DeprecationWarning)
            type = preprocess
        self._type = self._auto_type_process(type)

    def __get__(self, instance: 'BaseConfig', owner):

        # initialize value
        if self._should_assign_default_value(instance):
            if self._default is None and not self._nullable:
                raise AttributeError("attribute must assign the value before access it.")

            self.__set__(instance, self._default)

        return vars(instance)[self.name]

    def __set__(self, instance, value):
        if value is None:
            if not self._nullable:
                raise ValueError('the value should not be none')
        else:
            value = self._type(value)

        vars(instance)[self.name] = value

    def __delete__(self, instance):
        del vars(instance)[self.name]

    def __set_name__(self, owner, name):
        if self.name is None:
            self.__name__ = name

    def _should_assign_default_value(self, instance):
        return self.name not in vars(instance)

    def _auto_type_process(self, typ: Type) -> Callable:
        if self.auto_boolean_process and typ is bool:
            return boolean
        if self.auto_config_process:
            from ..config import BaseConfig  # lazy import
            if issubclass(typ, BaseConfig):
                return config(typ)
        return typ

    @property
    def name(self) -> str:
        return self.__name__

    @property
    def type(self) -> Type:
        return self._type

    @property
    def required(self) -> bool:
        return self._required

    @property
    def description(self) -> str:
        return self._description



from copy import deepcopy
from typing import TypeVar

from . import flag_string

T = TypeVar("T", list, dict)


def flag_container(container: T) -> T:
    result = deepcopy(container)
    opening = [result]

    while len(opening) >= 0:
        container = opening.pop()
        for k in container:
            if isinstance(container[k], (list, dict)):
                opening.append(container[k])
            else:
                container[k] = flag_string(container[k])
    return result



from fancy import config as cfg


class MyConfig(cfg.BaseConfig):

    _a: int = cfg.Option(name="a", required=True, description="hi", type=int)
    b: float = cfg.Option(required=False, type=float)

    def post_load(self):
        print(self.a)

    @property
    def a(self):
        return self._a + 1


def test_config():
    opt = {
        "a": "3",
        "b": "4.2"
    }

    c = MyConfig(cfg.DictConfigLoader(opt))
    print(vars(c))

